@@ -24,10 +24,10 @@ typedef union {
 # define	K_OPEN_RO	262
 # define	K_OPEN_RW	263
 # define	K_OPEN_AO	264
-# define	K_AUTH	265
-# define	K_RUNAS	266
-# define	K_FORK	267
-# define	K_EXEC	268
+# define	K_UNLINK	265
+# define	K_AUTH	266
+# define	K_RUNAS	267
+# define	K_FORK	268
 # define	K_UNPRIV_USER	269
 # define	K_CHROOT	270
 # define	K_ALLOW_RERUN	271
