@@ -6,7 +6,7 @@
  * An encapsulation layer for msghdr structures, cause that's the worst
  * system-level API I have ever seen.
  *
- * $Id: msghdr.c,v 1.14 2002/10/11 02:31:51 dougk Exp $
+ * $Id: msghdr.c,v 1.15 2002/10/15 23:20:48 dougk Exp $
  */
  
 #include "../config.h"
@@ -21,6 +21,7 @@
 #include <sys/uio.h>
 #include <sys/socket.h>
 #include <syslog.h>
+#include <errno.h>
 
 /*
  * 1: Data structure comments.
@@ -335,6 +336,34 @@ void msg_getString(message_t *msg, char *buffer, size_t bufferlen) {
     }
 }
 
+char *msg_getAllocStr(message_t *msg, size_t maxlen) {
+    char *retval;
+    size_t stringlen = msg_getInt(msg);
+
+    if (maxlen <= 0)
+        maxlen = 4096; /* Random default size */
+
+    /* the string is not null terminated, so make sure we have the space */
+    maxlen = MIN(maxlen-1, stringlen);
+
+    retval = malloc(maxlen+1);
+    if (retval == NULL) {
+        syslog(LOG_ERR,"msg_getAllocStr: No Mem");
+        errno = ENOMEM;
+        return NULL;
+    }
+    msg_getData(msg, retval, maxlen);
+    retval[maxlen] = '\0';
+    /* Throw away any remaining */
+    if (msg->offset > MSG_PAGE) {
+        msg->cur_iov += (msg->offset / MSG_PAGE);
+        msg->offset %= MSG_PAGE;
+        msg->recv_len -= stringlen;
+    }
+
+    return retval;
+}
+
 
 int msg_getFd  (message_t *msg)
 {
