@@ -6,12 +6,13 @@
  * Provides the client half of the process.  Should be considered
  * untrusted by the privman server.
  *
- * $Id: priv_client.cc,v 1.24 2002/10/10 19:17:14 dougk Exp $
+ * $Id: priv_client.cc,v 1.35 2002/10/31 16:11:39 dougk Exp $
  */
 
 #include "../config.h"
 #include <stdarg.h>
 #include <stdlib.h>
+#include <stdio.h>
 #include <strings.h>
 #include <string.h>
 #include <unistd.h>
@@ -23,6 +24,14 @@
 #include <assert.h>
 #include <sys/wait.h>
 #include <syslog.h>
+#include <sys/socket.h>
+#include <sys/un.h>
+
+#include <map>
+
+#ifdef HAVE_SYS_RESOURCE_H
+#include <sys/resource.h>
+#endif
 
 #include "privman.h"
 #include "msghdr.h"
@@ -30,6 +39,13 @@
 #include "types.h"
 #include "priv_impl.h"
 
+#ifndef UNIX_PATH_MAX
+#define UNIX_PATH_MAX 108
+#endif
+
+extern char **environ;
+
+#ifdef HAVE_LIBPAM
 /* Used as a cache for get/set_item */
 static const void      *pam_types[20] = { 0 };
 /*                                      // Strings.  void* for strings.
@@ -43,6 +59,7 @@ static const void      *pam_types[20] = { 0 };
  *  [PAM_USER_PROMPT]       = NULL,
  *  [PAM_FAIL_DELAY]        = NULL      // This is just a function pointer
  */
+#endif /* HAVE_LIBPAM */
  
 #ifndef CONFIG_PATH
 #define CONFIG_PATH="/etc/privman.d"
@@ -67,6 +84,7 @@ static void readConfig(const char *progname) {
         fclose(yyin);
 }
 
+#ifdef HAVE_LIBPAM
 /* When requested, calls the PAM conversion function registered by
  * the client.
  */
@@ -116,6 +134,8 @@ static void handleConvert(message_t *msg)
     free(resp);
 }
 
+#endif /* HAVE_LIBPAM */
+
 
 static __inline__
 void wait_for_debugger(void)
@@ -141,45 +161,44 @@ void socketfun(int sockfds[2], bool server) {
     }
 }
 
-void setup_child(int sockfds[2],
-        void (*fnptr)(const char *), const char *arg)
+void setup_child(void (*fnptr)(char * const *), char * const args[],
+        const char *user, const char *root)
 {
     struct passwd  *pwent;
 
-    wait_for_debugger();
-
-    /* in the child process */
-    socketfun(sockfds, false);
-
     /* Get unpriv_user info, in case chroot changes it.
      * chroot,
      * setuid.
      */
 
-    /* Normalize config. */
-    if (config->unpriv_user == "")
-        config->unpriv_user = "nobody";
-    if (config->unpriv_jail == "")
-        config->unpriv_jail = "/";
+    /* Normalize unpriv_user, root. */
+    if (user == NULL || (strcmp(user, "") == 0))
+        user = "nobody";
+    if (root == NULL || (strcmp(root, "") == 0))
+        root = "/";
 
     /* getpwnam */
-    pwent = getpwnam(config->unpriv_user.c_str());
+    pwent = getpwnam(user);
     /* Don't know if its a static pointer, or malloced and I'm allowed
      * to clear it, so just leak it.
      */
 
     if (pwent == NULL) {
-        syslog(LOG_ERR, "getpwnam failed on unpriv user %s",
-            config->unpriv_user.c_str());
+        syslog(LOG_ERR, "getpwnam failed on unpriv user %s", user);
         boom("setup_child(getpwnam)");
     }
 
     /* chroot */
-    if (chroot(config->unpriv_jail.c_str()) < 0) {
-        syslog(LOG_ERR, "chroot to %s\n", config->unpriv_jail.c_str());
+    if (chroot(root) < 0) {
+        syslog(LOG_ERR, "chroot to %s\n", root);
         boom("setup_child(chroot)");
     }
 
+    if (chdir("/") < 0) {
+        syslog(LOG_ERR, "chroot to %s\n", root);
+        boom("setup_child(chdir)");
+    }
+
     if (setgid(pwent->pw_gid) < 0)
         boom("setup_child(setgid)");
     if (setuid(pwent->pw_uid) < 0)
@@ -187,14 +206,17 @@ void setup_child(int sockfds[2],
 
     /* Call provilded function */
     if (fnptr != NULL) {
-        fnptr(arg);
+        fnptr(args);
     }
 
-    /* And return to do normal work. */
+    /* And return to do normal work.   Or not.*/
+    if (privmand_fd == -1)
+        _exit(0);
 }
 
 void priv_sep_init(void (*servfn)(void),
-    void (*childfn)(const char *), const char *childfn_arg)
+    void (*childfn)(char * const *), char * const childfn_args[],
+    const char *user, const char *root)
 {
     int         sockfds[2];
 
@@ -203,7 +225,10 @@ void priv_sep_init(void (*servfn)(void),
 
     child_pid = fork();
     if (child_pid == 0) {
-        setup_child(sockfds, childfn, childfn_arg);
+        wait_for_debugger();
+
+        socketfun(sockfds, false);
+        setup_child(childfn, childfn_args, user, root);
     } else if (child_pid < 0) {
         boom("fork");
     } else {
@@ -225,7 +250,8 @@ void priv_init(const char *appname)
     /* Read the config now. */
     readConfig(appname);
 
-    priv_sep_init(privman_serv_init, 0, 0);
+    priv_sep_init(privman_serv_init, 0, 0,
+            config->unpriv_user.c_str(), config->unpriv_jail.c_str());
 }
 
 
@@ -326,6 +352,48 @@ FILE* priv_fopen(const char *pathname, const char *mode)
    return fdopen(fd, mode);
 }
 
+int priv_unlink(const char *pathname)
+{
+    message_t          *msg = msg_new();
+    int                 n, retval;
+    char                cwd[PATH_MAX];
+
+    msg_init(msg, CMD_UNLINK);
+
+    /* We have to canpath the path, else chdir() messes us up. */
+    if (getcwd(cwd, sizeof(cwd)) == NULL) {
+        /* Use a token "NFC" value.  we won't work right in this
+         * case, but it might work if you didn't chdir
+         */
+        msg_addString(msg,"");
+    } else {
+        msg_addString(msg,cwd);
+    }
+    msg_addString(msg,pathname);
+    /* send the message */
+    n = msg_sendmsg(msg, privmand_fd);
+    if (n < 0) {
+        retval = -1;
+        goto exit;
+    }
+
+    /* listen for responce */
+    msg_clear(msg);
+    n = msg_recvmsg(msg, privmand_fd);
+    if ( n < 0 ) {
+        retval = -1;
+        goto exit;
+    }
+
+    retval = msg_getInt(msg);
+    if ( retval < 0 ) {
+        errno = -n;
+        retval = -1;
+    }
+exit:
+    msg_delete(msg);
+    return retval;
+}
 int priv_bind(int sockfd, struct sockaddr *addr, socklen_t addrlen)
 {
     message_t  *msg = msg_new();
@@ -337,9 +405,7 @@ int priv_bind(int sockfd, struct sockaddr *addr, socklen_t addrlen)
     msg_addData(msg, addr, addrlen);
 
     /* send the message */
-    n = msg_sendmsg(msg, privmand_fd);
-    if (n < 0)
-        boom("priv_bind(sendmsg)");
+    msg_sendmsg(msg, privmand_fd, "priv_bind(sendmsg)");
 
     /* listen for responce */
     msg_clear(msg);
@@ -354,6 +420,7 @@ int priv_bind(int sockfd, struct sockaddr *addr, socklen_t addrlen)
     return n;
 }
 
+#ifdef HAVE_LIBPAM
 int priv_pam_start(const char *service, const char *user,
                     const struct pam_conv *conv,
                     pam_handle_t **pamh_p)
@@ -394,6 +461,7 @@ int priv_pam_start(const char *service, const char *user,
         *pamh_p = (pam_handle_t*)msg_getPtr(msg);
     }
 
+    msg_delete(msg);
     return retval;
 }
 
@@ -437,6 +505,7 @@ static int priv_pam_simple_func(pam_handle_t *pamh, int flags,
         }
     } while (cmd != PRIV_PAM_RC);
 
+    msg_delete(msg);
     return rc;
 }
 
@@ -462,7 +531,9 @@ PRIV_PAM_SIMPLE(pam_setcred,            CMD_PAM_SETCRED)
 PRIV_PAM_SIMPLE(pam_chauthtok,          CMD_PAM_CHAUTHTOK)
 PRIV_PAM_SIMPLE(pam_open_session,       CMD_PAM_OPEN_SESSION)
 PRIV_PAM_SIMPLE(pam_close_session,      CMD_PAM_CLOSE_SESSION)
+#ifdef LO_HAVE_PAM_FAIL_DELAY
 PRIV_PAM_SIMPLE2(pam_fail_delay,        CMD_PAM_FAIL_DELAY)
+#endif
 
 int priv_pam_set_item(pam_handle_t *pamh, int item_type, const void *item)
 {
@@ -510,6 +581,7 @@ int priv_pam_set_item(pam_handle_t *pamh, int item_type, const void *item)
         }
     }
 
+    msg_delete(msg);
     return rc;
 }
 
@@ -558,6 +630,7 @@ int priv_pam_get_item(pam_handle_t *pamh, int item_type, const void **item)
 
         *item = pam_types[item_type];
     }
+    msg_delete(msg);
     return rc;
 }
 
@@ -575,6 +648,7 @@ PAM:
     pam_getenvlist(pam_handle)
 
 */
+#endif /* HAVE_LIBPAM */
 
 pid_t priv_fork(void)
 {
@@ -628,6 +702,7 @@ pid_t priv_fork(void)
     }
 
 exit:
+    msg_delete(msg);
     return retval;
 }
 
@@ -682,48 +757,86 @@ int priv_daemon(int nochdir, int noclose)
  * Will be executed with as user, so check permissions.
  *
  * Actually Executes from the parent, then exit's here.
+ * priv_rerunas(), with function pointer that demarshalls args, 
+ * tells the monitor to quit, and exec's.
  */
+static void priv_execve_impl(char * const arg[]);
 int priv_execve(const char *filename, char * const argv[], char * const envp[],
             const char *user, const char *root)
 {
-    message_t *msg = msg_new();
-    int i, argc, envc;
+    const char**arg;
+    int         i, j, argc, envc;
+    char        buf[5]; /* "9999\0" */
 
-    msg_init(msg, CMD_EXEC_AS);
-    msg_addString(msg, filename);
-    msg_addString(msg, user != NULL ? user : "");
-    msg_addString(msg, root != NULL ? root : "");
-
-    for (argc = 0; argv[argc] != NULL; ++argc)
+    for (argc = 0; argv[argc] != NULL && argc < 9999; ++argc)
         ;
-    msg_addInt(msg, argc);
-    for (i = 0; i < argc; ++i) {
-        msg_addString(msg, argv[i]);
+    for (envc = 0; envp[envc] != NULL && envc < 9999; ++envc)
+        ;
+
+    /* arg = { path + "3" + argv[0 .. 2] + "2" + envp[0 .. 1] + 0}; */
+    arg = (const char **)malloc(sizeof(char *) * (argc + envc + 2 + 1 + 1));
+
+    i = 0;
+    arg[i++] = filename;
+
+    snprintf(buf, sizeof(buf)-1, "%d", argc); buf[sizeof(buf)-1] = '\0';
+    arg[i++] = strdup(buf);
+
+    for ( j = i ; i < argc + j; ++i) {
+        arg[i] = argv[i-j];
     }
 
-    for (envc = 0; envp[envc] != NULL; ++envc)
-        ;
-    msg_addInt(msg, envc);
-    for (i = 0; i < envc; ++i) {
-        msg_addString(msg, envp[i]);
+    snprintf(buf, sizeof(buf)-1, "%d", envc); buf[sizeof(buf)-1] = '\0';
+    arg[i++] = strdup(buf);
+
+    for ( j = i ; i < envc + j; ++i) {
+        arg[i] = envp[i-j];
     }
+    arg[i] = NULL;
 
-    /* send the message */
-    msg_sendmsg(msg, privmand_fd, "priv_execve(sendmsg)");
+    i = priv_rerunas(priv_execve_impl, (char * const *)arg, user, root, 0);
 
-    /* listen for responce */
-    msg_clear(msg);
-    msg_recvmsg(msg, privmand_fd, "priv_execve(recvmsg)");
-    i = msg_getInt(msg);
     if (i < 0) {
-        errno = -i;
-        i = -1;
-    } else {
-        _exit(0);
+        free(arg);
+        return i;
     }
+    /* The exec should happen in the new slave.  We want the monitor
+     * to wait on the new child to exit so that it appears to exit when
+     * the new child does.
+     */
+    _exit(0);
+}
 
-    msg_delete(msg);
-    return i;
+static void priv_execve_impl(char * const arg[])
+{
+    const char         *filename;
+    char              **argv;
+    char              **envp;
+    int                 argc, envc, i, j;
+
+    /* Tell the monitor to exit */
+    priv_exit(0);
+
+    i = 0;
+    filename = arg[i++];
+
+    argc = atoi(arg[i++]);
+    argv = (char**)malloc(sizeof(char*) * (argc + 1));
+    for (j = 0; j < argc; ++j) {
+        argv[j] = arg[i++];
+    }
+    argv[j] = NULL;
+
+    envc = atoi(arg[i++]);
+    envp = (char**)malloc(sizeof(char*) * (envc + 1));
+    for (j = 0; j < envc; ++j) {
+        envp[j] = arg[i++];
+    }
+    envp[j] = NULL;
+
+    execve(filename, argv, envp);
+    perror("priv_execve_impl(execve)");
+    _exit(EXIT_FAILURE);
 }
 
 /* Creates a clean fork() from the privmand server with state equivilent
@@ -734,15 +847,16 @@ int priv_execve(const char *filename, char * const argv[], char * const envp[],
  * TODO: limits.  Feh.
  *
  */
-int priv_rerunas(void (*fnptr)(const char *), const char *arg,
-            const char *user, const char *root)
+int priv_rerunas(void (*fnptr)(char * const *), char * const arg[],
+            const char *user, const char *root, int flags)
 {
-    message_t *msg = msg_new();
     int i;
+    message_t *msg = msg_new();
 
     msg_init(msg, CMD_RERUN_AS);
+    msg_addInt(msg, flags);
     msg_addPtr(msg, (void*)fnptr);
-    msg_addString(msg, arg);
+    msg_addArgv(msg, arg);
     msg_addString(msg, user != NULL ? user : "");
     msg_addString(msg, root != NULL ? root : "");
 
@@ -756,12 +870,342 @@ int priv_rerunas(void (*fnptr)(const char *), const char *arg,
     if (i < 0) {
         errno = -i;
         i = -1;
-    } else {
-        /* _exit since the "real program" transitioned to the new one. */
-        _exit(0);
+    } else if (!(flags & PRIV_RR_OLD_SLAVE_MONITORED)) {
+        /* the "real program" transitioned to the new one. */
+        close(privmand_fd);
     }
 
     msg_delete(msg);
     return i;
 }
 
+/* Creates a clean fork() from a new privmand server with state equivilent
+ * to the client state when priv_init() was first called.
+ *
+ * Invokes the specified function, with the string arg provided.
+ * chroots to the directory provided, setuid's to the user provided.
+ * TODO: limits.  Feh.
+ *
+ */
+int priv_respawn_as(void (*fnptr)(char * const *), char * const arg[],
+            const char *user, const char *root)
+{
+    message_t *msg = msg_new();
+    int i;
+
+    msg_init(msg, CMD_RESPAWN_AS);
+    msg_addPtr(msg, (void*)fnptr);
+    msg_addArgv(msg, arg);
+    msg_addString(msg, user != NULL ? user : "");
+    msg_addString(msg, root != NULL ? root : "");
+
+    /* send the message */
+    msg_sendmsg(msg, privmand_fd, "priv_respawn_as(sendmsg)");
+
+    /* listen for responce */
+    msg_clear(msg);
+    msg_recvmsg(msg, privmand_fd, "priv_respawn_as(recvmsg)");
+    i = msg_getInt(msg);
+    if (i < 0) {
+        errno = -i;
+        i = -1;
+    }
+
+    msg_delete(msg);
+    return i;
+}
+/* Proxies wait4.  Nothing is really privileged about this call, but
+ * due to the way rerunas et. all work, sometimes the wait has to be
+ * done from the monitor.
+ */
+pid_t	priv_wait4(pid_t pid, int *status, int options, struct rusage *rus)
+{
+    message_t  *msg = msg_new();
+    pid_t       retval;
+    int         flags = 0;
+
+    if (status != NULL)
+        flags |= WANTS_STATUS;
+    if (rus    != NULL)
+        flags |= WANTS_RUSAGE;
+
+    msg_init(msg, CMD_WAIT4);
+
+    msg_addInt(msg, pid);
+    msg_addInt(msg, options);
+    msg_addInt(msg, flags);
+
+    msg_sendmsg(msg, privmand_fd, "priv_wait4(sendmsg)");
+    msg_clear(msg);
+
+    msg_recvmsg(msg, privmand_fd, "priv_wait4(recvmsg)");
+
+    retval = msg_getInt(msg);
+
+    if (retval < 0) {
+        errno = -retval;
+        retval = -1;
+    } else {
+        if (status != NULL)
+            *status = msg_getInt(msg);
+        if (rus    != NULL)
+            msg_getData(msg, rus, sizeof(*rus));
+    }
+
+    msg_delete(msg);
+    return retval;
+}
+
+
+/* Table needed for popen.  We have to be able to map a handle to 
+ * a File Descriptor.. because CMD_PCLOSE requires the handle that
+ * POPEN_AS returned so that it knows which process is expected to end.
+ *
+ * Maps FILE* to pid.
+ */
+static std::map<FILE*, int> file_pid_map;
+/* 
+ * popen, as the specified user in the specified chroot.
+ * Er, as this is on top of rerun, the level of ganularity of
+ * control may be inadaquate.
+ */
+static void priv_popen_impl(char * const arg[])
+{
+    int                 i, fd;
+    struct sockaddr_un  saddr;
+    char               *argv[] = { "sh", "-c", arg[0], 0 };
+    char               *type;
+
+    /* 1. Close all fd's except for stdin,out,err.                      */
+    for (i = 3; i < 255; ++i)
+        close(i);
+
+    /* 2. Connect to the unix domain socket provided.                   */
+    memset(&saddr, 0, sizeof(saddr));
+    saddr.sun_family = AF_LOCAL;
+    memcpy(saddr.sun_path, arg[2], strlen(arg[2])+1);
+
+    if ((fd = socket(AF_LOCAL, SOCK_STREAM, 0)) < 0) {
+        boom("priv_popen_as(socket)");
+    }
+    if (connect(fd, (struct sockaddr*)&saddr, sizeof(saddr)) < 0) {
+        boom("priv_popen_as(connect)");
+    }
+
+    /* 3. Dup the socket onto the fd requested.  Close the socket       */
+    type = arg[1];
+    if (type[0] == 'w') {
+        dup2(fd, STDIN_FILENO);
+    } else if (type[0] == 'r') {
+        dup2(fd, STDOUT_FILENO);
+    }
+    close(fd);
+
+    /* 4. execve.                                                       */
+    execve("/bin/sh", argv, environ);
+    _exit(-1);
+}
+
+
+FILE*   priv_popen_as(const char *command, const char *type, const char *user)
+{
+    char               *socketname;
+    int                 sock, i;
+    struct sockaddr_un  saddr;
+    pid_t               pid;
+    FILE               *f;
+
+    /* 0.  Stupid checks.                                               */
+    if (type[1] != '\0' || (type[0] != 'r' && type[0] != 'w')) {
+        errno = EINVAL;
+        return NULL;
+    }
+
+    /* 1.  Create a socket name random-nuf.
+     * Bad.  RACEABLE.  I'd use anonymous, but they don't appear to be
+     * universally supported.  This is why I don't support chroot.
+     */
+    
+    do {
+        if ((sock = socket(AF_LOCAL, SOCK_STREAM, 0)) < 0)
+            boom("priv_popen_as(socket)");
+
+        socketname = tempnam(0, "privmn");
+
+        memset(&saddr, 0, sizeof(saddr));
+        saddr.sun_family = AF_LOCAL;
+        memcpy(saddr.sun_path,socketname,strlen(socketname)+1);
+
+        i = bind(sock, (sockaddr*)&saddr, sizeof(saddr));
+
+        if (i < 0)
+            close(sock);
+
+    } while (i == -1 && errno == EINVAL);
+    if (i < 0)
+        boom("priv_popen_as(bind)");
+
+    /* And chmod it 666 */
+    chmod(saddr.sun_path, 0666);
+
+    if (listen(sock, 1) < 0)
+        boom("priv_popen_as(listen)");
+
+    /* 2.  Rerun_as.  Args are command, type, and filename.           */
+
+    char * const args[4] = { (char*)command, (char*)type, socketname, 0};
+
+    pid = priv_rerunas(priv_popen_impl, args, user, NULL,
+            PRIV_RR_OLD_SLAVE_MONITORED);
+
+    /* 3.  Get fd.  Close off socket since we are not listening anymore. */
+    int         fd;
+    if ((fd = accept(sock, 0, 0)) < 0) {
+        perror("accept failed.\n");
+        boom("priv_popen_as(accept)");
+    }
+
+    close(sock);
+    unlink(socketname);
+
+    /* 4.  Turn it into a FILE* and add to the map.                     */
+    f = fdopen(fd, type);
+    file_pid_map[f] = pid;
+
+    /* 5.  Done.                                                        */
+    return f;
+}
+
+int     priv_pclose(FILE *stream)
+{
+    int                 status, i;
+    /* 1. Get pid from map, and remove from map                         */
+    pid_t               pid = file_pid_map[stream];
+    file_pid_map.erase(stream);
+
+    /* 2. fclose() the stream.  This may cause the process to keel over
+     * but that's a good thing.                                         */
+    fclose(stream);
+
+    /* 3. Collect the status.  Its a grandchild of our parent so having
+     * the parent wait for it should work.                              */
+    do {
+        i = priv_wait4(pid, &status, 0, NULL);
+    } while (i == -1 && errno == EINTR);
+
+    return status;
+}
+
+
+/*
+ * Extension framework. 
+ */
+
+info_fn_map_t info_fn_map;
+ cap_fn_map_t  cap_fn_map;
+
+/* Prevent the same handle from being in both maps. */
+static int handle_counter = 0;
+
+int priv_register_info_fn(char *(*fnptr)(char * const *))
+{
+    int handle;
+ 
+    /* Must be done before the call to priv_init() */
+    if (child_pid == 0) {
+        errno = EPERM;
+        return -1;
+    }
+
+    handle = handle_counter++;
+
+    info_fn_map[handle] = fnptr;
+    return handle;
+}
+
+int priv_register_cap_fn(int (*fnptr)(char * const *))
+{
+    int handle;
+ 
+    /* Must be done before the call to priv_init() */
+    if (child_pid == 0) {
+        errno = EPERM;
+        return -1;
+    }
+
+    handle = handle_counter++;
+
+    cap_fn_map[handle] = fnptr;
+    return handle;
+}
+
+
+char *priv_invoke_info_fn(int handle, char * const args[])
+{
+    message_t  *msg = msg_new();
+    char       *retval;
+    int         rc;
+
+    msg_init(msg, CMD_CUSTOM_INFO);
+
+    msg_addInt(msg, handle);
+    msg_addArgv(msg, args);
+
+    /* send request */
+    msg_sendmsg(msg, privmand_fd, "priv_invoke_info_fn(sendmsg)");
+    /* listen for responce */
+    msg_clear(msg);
+    msg_recvmsg(msg, privmand_fd, "priv_invoke_info_fn(recvmsg)");
+
+    rc = msg_getInt(msg);
+    if (rc < 0) {
+        errno = -rc;
+        retval = NULL;
+    } else {
+        retval = msg_getAllocStr(msg, 4096);
+    }
+    msg_delete(msg);
+    return retval;
+}
+
+int priv_invoke_cap_fn(int handle, char * const args[])
+{
+    message_t  *msg = msg_new();
+    int         rc;
+
+    msg_init(msg, CMD_CUSTOM_INFO);
+
+    msg_addInt(msg, handle);
+    msg_addArgv(msg, args);
+
+    /* send request */
+    msg_sendmsg(msg, privmand_fd, "priv_invoke_cap_fn(sendmsg)");
+    /* listen for responce */
+    msg_clear(msg);
+    msg_recvmsg(msg, privmand_fd, "priv_invoke_cap_fn(recvmsg)");
+
+    rc = msg_getInt(msg);
+    if (rc < 0) {
+        errno = -rc;
+        rc = -1;
+    } else {
+        rc = msg_getFd(msg);
+    }
+    msg_delete(msg);
+    return rc;
+}
+
+/* Doesn't exit, just causes the privman server to. */
+void priv_exit(int status)
+{
+    message_t  *msg;
+
+    msg = msg_new();
+
+    msg_init(msg, CMD_EXIT);
+    msg_addInt(msg, status);
+    msg_sendmsg(msg, privmand_fd); /* Never check an error you can't handle */
+    close(privmand_fd);
+}
+
+
