@@ -6,7 +6,7 @@
  * Provides the server half of the process.  This half retains priviledge
  * and should be nei-invulnerable.
  *
- * $Id: privman.cc,v 1.37 2002/10/10 19:17:14 dougk Exp $
+ * $Id: privman.cc,v 1.50 2002/10/31 16:11:39 dougk Exp $
  */
 #include "../config.h"
 #include <stdio.h>
@@ -27,10 +27,19 @@
 #ifdef HAVE_STRINGS_H
 #include <strings.h>
 #endif
-#ifdef HAVE_SYS_BYTEORDER_H
-#include <sys/byteorder.h>
+
+#ifdef HAVE_ARPA_INET_H
+#include <arpa/inet.h>
+#endif
+#ifdef HAVE_NETINET_IN_H
+#include <netinet/in.h>
 #endif
 
+#ifdef HAVE_SYS_RESOURCE_H
+#include <sys/resource.h>
+#endif
+
+#include <set>
 #include <map>
 
 #include "privman.h"
@@ -43,9 +52,6 @@ int             privmand_fd     = -1;    /* FD to talk to the other process */
 
 pid_t           child_pid       = 0;
 
-/* TBD: multiple conversion functions. */
-static       struct pam_conv    pconv;    /* server side. */
-
 static bool     p_wait_on_child = true;
 
 /* The config for this invocation.  Not static so that the parser can
@@ -54,93 +60,62 @@ static bool     p_wait_on_child = true;
 config_t               *config;
 
 /* Dispatch table */
-std::map< enum commands, void(*)(message_t*) > function_map;
+static std::map< enum commands, void(*)(message_t*) > function_map;
 
-/* PAM conversion function.  We pass the work back to the other
- * side.  The assumption is that the other side is currently
- * waiting in a priv_pam_foo() call, and will recoginize the
- * message.
- *
- * Don't bother passing app_ptr, as it won't be right anyway.  The other
- * side has it.  resp is an out value, not an in value.
- */
-
-#ifdef __cplusplus
-#define UNUSED(p)
-#else
-#define UNUSED(p) p __attribute((unused))
-#endif
-
-static int convert_punt(int num_msg,
-        PAM_CONV_FUNC_CONST struct pam_message **messages,
-        struct pam_response **resp,
-        void * UNUSED(app_ptr))
+static inline
+void freeArgv(char * argv[])
 {
-    message_t                  *msg = msg_new();
-    struct pam_response        *reply;
-    int                         retval, i;
-
-    msg_initResponce(msg, PRIV_PAM_RUN_CONV);
-    msg_addInt(msg,num_msg);
-    for (i = 0; i < num_msg; ++i) {
-        msg_addInt(msg,messages[i]->msg_style);
-        msg_addString(msg,messages[i]->msg);
-    }
-
-    msg_sendmsg(msg, privmand_fd, "convert_punt(sendmsg)");
-
-    msg_initResponce(msg, 0);
-    msg_recvmsg(msg, privmand_fd, "convert_punt(recvmsg)");
-
-    retval = msg_getInt(msg);
-    reply = (struct pam_response*)malloc(sizeof(*reply) * num_msg);
-    for (i = 0; i < num_msg; ++i) {
-        char    buf[PAM_MAX_RESP_SIZE];
-        msg_getString(msg, buf, sizeof(buf)-1);
-        buf[sizeof(buf)-1] = '\0';
-
-        reply[i].resp = strdup(buf);
-        reply[i].resp_retcode = msg_getInt(msg);
-    }
-
-    msg_delete(msg);
-
-    *resp = reply;
-    return retval;
+    for (int i = 0; argv[i] != NULL; ++i)
+        free(argv[i]);
+    free(argv);
 }
 
-static bool openPerm(const char *path, int flags)
+static inline
+char *msg_getAllocStr(message_t *msg, size_t maxlen, const char *errormsg) {
+    char *rv = msg_getAllocStr(msg, maxlen);
+    if (rv == NULL)
+        boom(errormsg);
+    return rv;
+}
+
+typedef enum {
+    at_none             = -1,
+    at_read_only,
+    at_read_write,
+    at_append_only,
+    at_unlink
+} accessType_t;
+
+static accessType_t openAccessType(int flags) {
+    if ((flags & 3) == O_RDONLY)
+        return at_read_only;
+    else if (((flags & 3) == O_WRONLY) && (flags & O_APPEND))
+        return at_append_only;
+    else if (((flags & 3) == O_RDWR) || (flags & 3) == O_WRONLY)
+        return at_read_write;
+    else
+        return at_none;
+}
+
+static bool openPerm(const char *path, accessType_t type)
 {
     /* MAC check here. TBD: use real globs instead of this hack */
     /* Hack:
      *   1) Path is in set.  Go home happy.
      *   2) /first/bit/of/path/[*] is in set.  Go home happy.
      */
-    enum {
-        /* Use this enum to index into the list[] array below. */
-                            read_only       = 0,
-                            read_write      = 1,
-                            append_only     = 2,
-                            exec            = 3
-    }                   type;
+    /* Use the "type" enum to index into the list[] array. */
     path_list          *list[] = {
                             &(config->open_ro),
                             &(config->open_rw),
                             &(config->open_ao),
-                            &(config->exec_list)
+                            &(config->unlink)
                         };
     char                testpath[MAXPATHLEN+1];
     char               *offset;
 
-    /* Determine the type of access. */
-    if ((flags & 3) == O_RDONLY)
-        type = read_only;
-    else if (((flags & 3) == O_WRONLY) && (flags & O_APPEND))
-        type = append_only;
-    else if (((flags & 3) == O_RDWR) || (flags & 3) == O_WRONLY)
-        type = read_write;
-    else
-        type = exec;
+    if (type == at_none)
+        return false;
 
     strncpy(testpath, path, sizeof(testpath)-2);
     testpath[sizeof(testpath)-2] = '\0';
@@ -166,6 +141,23 @@ static bool openPerm(const char *path, int flags)
     return true;
 }
 
+static bool runasPerm(const char *user)
+{
+    if (user[0] != '\0' && config->user.count(user) == 0
+            && config->user.count("*") == 0)
+        return false;
+    return true;
+}
+
+static void sendEPERM(message_t *msg, const char *reason)
+{
+    msg_initResponce(msg, -EPERM);
+    if (reason != NULL)
+        syslog(LOG_NOTICE, "%s", reason);
+    msg_sendmsg(msg, privmand_fd, "sendEPERM(sendmsg)");
+}
+
+
 /* Like "realpath", but can cope with a missing file in a non
  * missing directory.  Might not handle dangling symlinks right. XXX
  */
@@ -208,32 +200,17 @@ static bool myrealpath(const char *path, char *resolved)
 
     return true;
 }
-/* Handle an "open file" request */
-static void openFile(message_t *msg) {
-    char        path[MAXPATHLEN+1];
-    char        cwd[MAXPATHLEN+1];
 
+/* Handle certain requests */
+
+static void unlinkFile(message_t *msg) {
+    char       *path;
+    char       *cwd;
     char        canpath[MAXPATHLEN+1];
-    int         flags, mode;
-    int         rfd = 0;
-    int         n;
+    int         retval, n;
 
-    /* Open file specified, return the fd. */
-    /* arg1 = open mode
-     * arg2 = creation mode (optional)
-     * arg3 = pathname
-     *
-     * Returns
-     * arg1 = 0 or -1
-     * arg2 = errno (if arg1 < 0)
-     */
-
-    flags = msg_getInt(msg);
-    mode  = msg_getInt(msg);
-    msg_getString(msg, cwd, sizeof(cwd)-1);
-    msg_getString(msg, path, sizeof(path)-1);
-    path[sizeof(path)-1] = '\0';
-    cwd[sizeof(cwd)-1] = '\0';
+    cwd   = msg_getAllocStr(msg, MAXPATHLEN+1, "unlinkFile: bad cwd");
+    path  = msg_getAllocStr(msg, MAXPATHLEN+1, "unlinkFile, path path");
 
     /* Canacolize the path, so that a comparison with the
      * allowed list makes sence, and in case the client did a chdir.
@@ -257,14 +234,81 @@ static void openFile(message_t *msg) {
          * realpath said.
          */
         msg_initResponce(msg, -errno);
+        msg_sendmsg(msg, privmand_fd, "unlinkFile(sendmsg)");
     }
 
-    if (!openPerm(canpath, flags)) {
-        /* MAC check here. TBD: use real globs instead of this hack */
-        msg_initResponce(msg, -EPERM);
-        syslog(LOG_NOTICE, "Unauthorized attempt open(%s, %d)", canpath, flags);
+    /* MAC check here. TBD: use real globs instead of this hack */
+    if (!openPerm(canpath, at_unlink)) {
+        sendEPERM(msg, "Unauthorized attempt to unlink");
     } else {
+        /* Ok, now do it. */
+        retval = unlink(canpath);
+        if (retval < 0) {
+            msg_initResponce(msg, -errno);
+            syslog(LOG_WARNING, "priv_unlink(unlink): %m");
+        } else {
+            msg_initResponce(msg, 0);
+        }
+        msg_sendmsg(msg, privmand_fd, "unlinkFile(sendmsg)");
+    }
 
+    free(path); free(cwd);
+}
+
+/* Handle an "open file" request */
+static void openFile(message_t *msg) {
+    char       *path;
+    char        cwd[MAXPATHLEN+1]; /* Has to be this big. */
+
+    char        canpath[MAXPATHLEN+1];
+    int         flags, mode;
+    int         rfd = 0;
+    int         n;
+
+    /* Open file specified, return the fd. */
+    /* arg1 = open mode
+     * arg2 = creation mode (optional)
+     * arg3 = pathname
+     *
+     * Returns
+     * arg1 = 0 or -1
+     * arg2 = errno (if arg1 < 0)
+     */
+
+    flags = msg_getInt(msg);
+    mode  = msg_getInt(msg);
+    msg_getString(msg, cwd, MAXPATHLEN+1);
+    path  = msg_getAllocStr(msg, MAXPATHLEN+1, "openFile, path path");
+
+    /* Canacolize the path, so that a comparison with the
+     * allowed list makes sence, and in case the client did a chdir.
+     *
+     * The client sends us "getcwd()" output, so no trailing '/'.  If
+     * the client wants to lie to us for this message, no great foul.
+     * He gets no file.
+     */
+
+    n = strlen(cwd);
+    if (path[0] == '/') { /* absolute or not? */
+        /* Abs: nuke cwd with path, then realpath it. */
+        strncpy(cwd, path, sizeof(cwd) - n);
+    } else {
+        cwd[n++] = '/'; /* Path seperator */
+        strncpy(cwd + n, path, sizeof(cwd) - n);
+    }
+
+    if (!myrealpath(cwd, canpath)) {
+        /* Could be lots of reasons.  So lets confuse em with whatever
+         * realpath said.
+         */
+        msg_initResponce(msg, -errno);
+        msg_sendmsg(msg, privmand_fd, "openFile(sendmsg)");
+    }
+
+    /* MAC check here. TBD: use real globs instead of this hack */
+    if (!openPerm(canpath, openAccessType(flags))) {
+        sendEPERM(msg, "Unauthorized attempt open");
+    } else {
         /* Ok, now do it. */
         rfd = open(canpath,flags,mode);
         if (rfd < 0) {
@@ -274,11 +318,11 @@ static void openFile(message_t *msg) {
             msg_initResponce(msg, 0);
             msg_setFd(msg,rfd);
         }
+        msg_sendmsg(msg, privmand_fd, "openFile(sendmsg)");
+        close(rfd); /* prevent the leak */
     }
 
-    msg_sendmsg(msg, privmand_fd, "openFile(sendmsg)");
-    if (rfd)
-        close(rfd); /* prevent the leak */
+    free(path);
 }
 
 static void bindPort(message_t *msg) {
@@ -299,9 +343,7 @@ static void bindPort(message_t *msg) {
     || ( config->bind_port.count(ntohs(addr->sin_port)) == 0
      &&  config->bind_port.count(ntohs(0)) == 0 ) ) { /* wildcard */
         /* Permission denied */
-        msg_initResponce(msg, -EPERM);
-        syslog(LOG_NOTICE, "Unauthorized attempt bind to port %d",
-                addr->sin_port);
+        sendEPERM(msg, "Unauthorzed attempt to bind to port.");
     } else {
         retval      = bind(sockfd, (struct sockaddr*)addr, addrlen);
 
@@ -309,23 +351,76 @@ static void bindPort(message_t *msg) {
             retval = -errno; /* since it'll be used for this anyway. */
 
         msg_initResponce(msg, retval);
+        msg_sendmsg(msg, privmand_fd, "bindPort(sendmsg)");
     }
 
-    msg_sendmsg(msg, privmand_fd, "bindPort(sendmsg)");
     close(sockfd); /* don't leak this. */
 }
-    
+ 
+#ifdef HAVE_LIBPAM
+
+/* TBD: multiple conversion functions. */
+static       struct pam_conv    pconv;    /* server side. */
+
+/* PAM conversion function.  We pass the work back to the other
+ * side.  The assumption is that the other side is currently
+ * waiting in a priv_pam_foo() call, and will recoginize the
+ * message.
+ *
+ * Don't bother passing app_ptr, as it won't be right anyway.  The other
+ * side has it.  resp is an out value, not an in value.
+ */
+
+#ifdef __cplusplus
+#define UNUSED(p)
+#else
+#define UNUSED(p) p __attribute((unused))
+#endif
+
+static int convert_punt(int num_msg,
+        PAM_CONV_FUNC_CONST struct pam_message **messages,
+        struct pam_response **resp,
+        void * UNUSED(app_ptr))
+{
+    message_t                  *msg = msg_new();
+    struct pam_response        *reply;
+    int                         retval, i;
+
+    msg_initResponce(msg, PRIV_PAM_RUN_CONV);
+    msg_addInt(msg,num_msg);
+    for (i = 0; i < num_msg; ++i) {
+        msg_addInt(msg,messages[i]->msg_style);
+        msg_addString(msg,messages[i]->msg);
+    }
+
+    msg_sendmsg(msg, privmand_fd, "convert_punt(sendmsg)");
+
+    msg_initResponce(msg, 0);
+    msg_recvmsg(msg, privmand_fd, "convert_punt(recvmsg)");
+
+    retval = msg_getInt(msg);
+    reply = (struct pam_response*)malloc(sizeof(*reply) * num_msg);
+    for (i = 0; i < num_msg; ++i) {
+        reply[i].resp = msg_getAllocStr(msg, PAM_MAX_RESP_SIZE,
+                "convert_punt: bad responce");
+        reply[i].resp_retcode = msg_getInt(msg);
+    }
+
+    msg_delete(msg);
+
+    *resp = reply;
+    return retval;
+}
+
 static void pamStart(message_t *msg)
 {
     pam_handle_t       *handle;
     int                 retval;
-    char                service[128];
-    char                user[128];
+    char               *service;
+    char               *user;
 
-    msg_getString(msg, service, sizeof(service)-1);
-    service[sizeof(service)-1] = '\0';
-    msg_getString(msg, user, sizeof(user)-1);
-    user[sizeof(user)-1] = '\0';
+    service = msg_getAllocStr(msg, 128, "pamStart: bad service");
+    user    = msg_getAllocStr(msg, 128, "pamStart: bad user");
 
     pconv = (struct pam_conv){ convert_punt, 0};
 
@@ -336,6 +431,7 @@ static void pamStart(message_t *msg)
     msg_addPtr(msg, handle);
 
     msg_sendmsg(msg, privmand_fd, "pamStart(sendmsg)");
+    free(service); free(user);
 }
 
 static void pamSimpleFunc(message_t *msg, int (*func)(pam_handle_t*,int))
@@ -410,28 +506,71 @@ static void pamGetItem(message_t *msg)
 
     msg_sendmsg(msg, privmand_fd, "pamGetItem(sendmsg)");
 }
+#endif
 
 static void exitServer(message_t *UNUSED(msg)) {
-    /* Single. */
-    _exit(0);
+    if (p_wait_on_child) {
+        /* This will cause control_loop to break out of the loop */
+        close(privmand_fd);
+    } else {
+        /* Simple. */
+        _exit(0);
+    }
+}
+
+static void privWait4(message_t *msg) {
+    pid_t               pid;
+    int                 options;
+    int                 status;
+    struct rusage       ruse;
+
+    int                *s = NULL;
+    struct rusage      *r = NULL;
+    int                 flags;
+
+    pid       = msg_getInt(msg);
+    options   = msg_getInt(msg);
+    flags     = msg_getInt(msg);
+    if (flags & WANTS_STATUS)
+        s = &status;
+    if (flags & WANTS_RUSAGE)
+        r = &ruse;
+
+    pid = wait4(pid, s, options, r);
+
+    msg_clear(msg);
+
+    if (pid < 0)
+        msg_initResponce(msg, -errno);
+    else {
+        msg_initResponce(msg, pid);
+
+        if (flags & WANTS_STATUS)
+            msg_addInt(msg, status);
+        if (flags & WANTS_RUSAGE)
+            msg_addData(msg, &ruse, sizeof(ruse));
+    }
+
+    msg_sendmsg(msg, privmand_fd, "privWait4(sendmsg)");
 }
 
 static void forkProcess(message_t *msg) {
     /* Client requests fork().
-     * We create new pid, hand back to client.
+     * We create new pipe fd, hand back to client.
      *    Fork.
-     *    Child privmand listens on pid.
-     *    We close pid.
+     *    Child privmand listens on fd.
+     *    We close fd.
+     *
+     * Can't quite do the same as we do in respawnAs, (and thus share code)
+     * as the child doesn't exit, it fork()s.
      */
     int         fds[2], n;
 
     if (socketpair(AF_LOCAL, SOCK_STREAM, 0, fds) < 0)
         boom("forkProcess(socketpair)");
 
-    msg_clear(msg);
-    msg_addInt(msg, 0); // Return code
+    msg_initResponce(msg, 0);
     msg_setFd(msg, fds[0]);
-
     msg_sendmsg(msg, privmand_fd, "forkProcess(sendmsg)");
 
     close(fds[0]);
@@ -450,7 +589,7 @@ static void forkProcess(message_t *msg) {
     }
 }
 
-void daemonProcess(message_t *msg) {
+static void daemonProcess(message_t *msg) {
     /* Daemon.
      * 1) Fork, parent quits.  No mucking with global state
      *    and the child should be fine.
@@ -460,7 +599,7 @@ void daemonProcess(message_t *msg) {
      *    I go away.  So set the p_wait_on_child flag.
      */
     int n;
-
+ 
     n = fork();
     if (n == 0) {
         /* child.  See above. */
@@ -485,178 +624,167 @@ void daemonProcess(message_t *msg) {
     }
 }
 
-void rerunAsProcess(message_t *msg)
+static void rerunAsProcess(message_t *msg)
 {
-    char        buf[4096];
-    char       *args;
-    char       *user;
-    char       *root;
-    void        (*fnptr)(const char*);
+    /* Flags tells us which child is allowed to talk to the privmand
+     * server.
+     */
+    char      **args;
+    char       *user, *root;
+    void        (*fnptr)(char * const *);
+    int         flags;
 
-    fnptr = (void (*)(const char*))msg_getPtr(msg);
-
-    msg_getString(msg, buf, sizeof(buf)-1);
-    buf[sizeof(buf)-1] = '\0';
-    args = strdup(buf);
-
-    msg_getString(msg, buf, sizeof(buf)-1);
-    buf[sizeof(buf)-1] = '\0';
-    user = strdup(buf);
-
-    msg_getString(msg, buf, sizeof(buf)-1);
-    buf[sizeof(buf)-1] = '\0';
-    root = strdup(buf);
+    flags = msg_getInt(msg);
+    fnptr = (void (*)(char* const *))msg_getPtr(msg);
+    args  = msg_getArgv(msg);
+    user  = msg_getAllocStr(msg, 32, "rerunAsProcess: bad user");
+    root  = msg_getAllocStr(msg, MAXPATHLEN+1, "rerunAsProcess: bad root");
 
     /* MAC checks. */
-    if (user[0] != '\0' && config->user.count(user) == 0
-            && config->user.count("*") == 0) {
-        msg_initResponce(msg, -EPERM);
-        syslog(LOG_NOTICE, "Unauthorized attempt rerunAs as user %s",
-                user);
+    if (!runasPerm(user)) {
+        sendEPERM(msg, "Unauthorized respawnAs");
+        goto cleanup;
+    }
+
+    if (flags & PRIV_RR_OLD_SLAVE_MONITORED) {
+        pid_t child2 = fork();
+        if (child2 == -1) {
+            boom("respawnAsProcess(fork2)");
+        } else if (child2 == 0) {
+            close(privmand_fd); /* cut connection to original privmand. */
+
+            setup_child(fnptr, args, user, root);
+        } else {
+            msg_initResponce(msg, child2);
+            msg_sendmsg(msg, privmand_fd, "respawnAsProcess(sendmsg)");
+        }
+    } else {
+        /* Tell the original slave to go away. */
+        msg_clear(msg);
+        msg_initResponce(msg, 0);
         msg_sendmsg(msg, privmand_fd, "rerunAsProcess(sendmsg)");
 
-        free(user);
-        free(root);
-        free(args);
-        return;
+        config->unpriv_user = user; /* Assignments, not copies. */
+        config->unpriv_jail = root; /* Just for consistency sake */
+        /* This will run the client function, then drop back
+         * through here, as both the parent and the child.
+         */
+        priv_sep_init(0, fnptr, args, user, root);
     }
 
-    /* config hacking.  Sanity checks? */
-    config->unpriv_user = user;
-    config->unpriv_jail = root;
-
-    /* the previous statement was actually a copy.  config-> has
-     * stl strings, not char * pointers.
-     */
+cleanup:
     free(user); free(root);
+    freeArgv(args);
+}
 
-    msg_clear(msg);
-    msg_initResponce(msg, 0);
-    msg_sendmsg(msg, privmand_fd, "rerunAsProcess(sendmsg)");
+static void respawnAsProcess(message_t *msg)
+{
+    char      **args;
+    char       *user, *root;
+    void        (*fnptr)(char * const *);
 
-    /* And if we wait on the child, wait here. */
-    if (child_pid != 0 && p_wait_on_child == true)
-        waitpid(child_pid, 0, 0);
+    fnptr = (void (*)(char* const *))msg_getPtr(msg);
+    args = msg_getArgv(msg);
+    user = msg_getAllocStr(msg, 32, "rerunAsProcess: bad user");
+    root = msg_getAllocStr(msg, MAXPATHLEN+1, "rerunAsProcess: bad root");
 
+    /* MAC checks. */
+    if (!runasPerm(user)) {
+        sendEPERM(msg, "Unauthorized respawnAs");
+        goto clean;
+    }
+
+    /* Fork the monitor.   Parent returns.  The rest of this method takes
+     * place in the child.
+     */
+    switch(fork()) {
+    case -1:
+        boom("respawnAsProcess(fork)");
+        /* NOTREACHED */
+    default:
+        /* Parent. */
+        msg_initResponce(msg, 0);
+        msg_sendmsg(msg, privmand_fd, "respawnAsProcess(sendmsg)");
+        goto clean;
+    case 0:
+        /* Child. Fall through. */
+        /* Close off connection with old monitor & old slave */
+        close(privmand_fd);
+        privmand_fd = -1;
+        break;
+    }
+
+    config->unpriv_user = user; /* Assignments, not copies. */
+    config->unpriv_jail = root; /* Just for consistency sake */
     /* This will run the client function, then drop back
      * through here, as both the parent and the child.
-     * Parent case is handled: globals have been changed, loop
-     * continues as before.
-     * Child side's a little different.
      */
-    priv_sep_init(0, fnptr, args);
-
-    free(args);
+    priv_sep_init(0, fnptr, args, user, root);
+clean:
+    free(user); free(root);
+    freeArgv(args);
 }
 
-void execAsProcess(message_t *msg) 
+/* populated by priv_register_foo in priv_client.cc.  Our copy should
+ * be protected by the fork() in priv_init, so after that it should
+ * not change.
+ */
+
+static void customInfo(message_t *msg)
 {
-    /* filename, user, root, argc, argv, envc, envp */
-    char program[MAXPATHLEN];
-    char username[32];
-    char rootpath[MAXPATHLEN];
-    int argc, envc;
-    char **argv;
-    char **envp;
+    int         handle  = msg_getInt(msg);
+    char      **args    = msg_getArgv(msg);
+    char       *rv      = NULL;
+    info_fn_map_t::iterator    it;
 
-    struct passwd     * pw;
-    struct stat         buf;
-    int                 i;
-#if 0
-    int                 orig_rootfd;
-#endif
-
-    msg_getString(msg, program,  sizeof(program) -1);
-    program[sizeof(program) -1] = '\0';
-    msg_getString(msg, username, sizeof(username)-1);
-    program[sizeof(username)-1] = '\0';
-    msg_getString(msg, rootpath, sizeof(rootpath)-1);
-    program[sizeof(rootpath)-1] = '\0';
-
-    /* Config file MAC */
-    if (!openPerm(program, 3)) {
-        msg_initResponce(msg, -EPERM);
-        syslog(LOG_NOTICE, "Unauthorized attempt execAs %s", program);
-        msg_sendmsg(msg, privmand_fd, "execAsProcess(sendmsg)");
-        return;
-    }
-    if (username[0] != '\0' && config->user.count(username) == 0
-            && config->user.count("*") == 0) {
-        msg_initResponce(msg, -EPERM);
-        syslog(LOG_NOTICE, "Unauthorized attempt execAs %s", username);
-        msg_sendmsg(msg, privmand_fd, "execAsProcess(sendmsg)");
-        return;
+    it = info_fn_map.find(handle);
+    if (it != info_fn_map.end()) {
+        rv = ((*it).second)(args);
+    } else {
+        errno = ENOENT;
     }
 
-    /* Check things out first.
-     * get "/" fd,
-     * chroot root
-     * stat program
-     * sanity check.
-     * fchdir rootfd
-     * chroot "."
-     */
-
-#if 0
-    if (rootpath != NULL)
-        orig_rootfd = open("/", O_DIRECTORY|O_RDONLY);
-#endif
-
-    /* XXX Return error. */
-    if (rootpath[0] != '\0' && chroot(rootpath) < 0) {
-        boom("execAsProcess(chroot)");
+    msg_initResponce(msg, PRIV_PAM_RC);
+    if (rv == NULL) {
+        msg_addInt(msg, -errno);
+    } else {
+        msg_addInt(msg, 0);
+        msg_addString(msg, rv);
     }
 
-    if (stat(program, &buf) < 0) {
-        boom("execAsProcess(stat)");
-    }
+    msg_sendmsg(msg, privmand_fd, "customInfo(sendmsg)");
 
-    if (!S_ISREG(buf.st_mode) || (
-                ! (buf.st_mode & S_IXOTH|S_IXUSR|S_IXGRP))) {
-        boom("execAsProcess(perms)");
-    }
-
-#if 0
-    fchdir(orig_rootfd);
-#endif
-    if (username[0] == '\0')
-        memcpy(username, "nobody", sizeof("nobody"));
-
-    pw = getpwnam(username);
-
-    if (pw == NULL)
-        boom("execAsProcess(getpwnam)");
-
-    setuid(pw->pw_uid);
-    setgid(pw->pw_gid);
-    /* bring in argc, envp, and exec */
-
-    argc = msg_getInt(msg);
-    argv = (char **) malloc(argc + 1 * sizeof(char*));
-    for ( i = 0; i < argc; ++i) {
-        char b[1024];
-        msg_getString(msg, b, sizeof(b)-1);
-        argv[i] = strdup(b);
-    }
-    argv[i] = NULL;
-
-    envc = msg_getInt(msg);
-    envp = (char **) malloc(envc + 1 * sizeof(char*));
-    for ( i = 0; i < envc; ++i) {
-        char b[1024];
-        msg_getString(msg, b, sizeof(b)-1);
-        envp[i] = strdup(b);
-    }
-    envp[i] = NULL;
-
-    msg_initResponce(msg, 0);
-    msg_sendmsg(msg, privmand_fd, "execAsProcess(sendmsg)");
-
-    execve(program, argv, envp);
-
-    boom("execAsProcess(execve)");
+    freeArgv(args);
+    free(rv);
 }
 
+static void customCap(message_t *msg)
+{
+    int         handle  = msg_getInt(msg);
+    char      **args    = msg_getArgv(msg);
+    int         rv;
+    cap_fn_map_t::iterator     it;
+
+    it = cap_fn_map.find(handle);
+    if (it != cap_fn_map.end()) {
+        rv = ((*it).second)(args);
+    } else {
+        rv = 0;
+        errno = ENOENT;
+    }
+
+    msg_initResponce(msg, PRIV_PAM_RC);
+    if (rv < 0) {
+        msg_addInt(msg, -errno);
+    } else {
+        msg_addInt(msg, 0);
+        msg_setFd(msg, rv);
+    }
+
+    msg_sendmsg(msg, privmand_fd, "customCap(sendmsg)");
+
+    freeArgv(args);
+}
 
 /* Is this client allowed to make this TYPE of request.  For
  * things with finer grained permissions, do that kind of check
@@ -666,10 +794,15 @@ static bool validRequest(enum commands c) {
     if (config == NULL)
         return false;
     switch (c) {
-    case CMD_OPEN:
-    case CMD_EXEC_AS:
-        return true; /* just deal with it later */
+    case CMD_OPEN:              /* Deal with it when we know what. */
+    case CMD_UNLINK:            /* Deal with it when we know what. */
+    case CMD_EXIT:              /* always thus. */
+    case CMD_WAIT4:             /* Seems harmless */
+    case CMD_CUSTOM_INFO:       /* job of other half to validate. */
+    case CMD_CUSTOM_CAP:
+        return true; 
     case CMD_RERUN_AS:
+    case CMD_RESPAWN_AS:
         return config->rerunas;
     case CMD_BIND:
         return !config->bind_port.empty();
@@ -687,10 +820,7 @@ static bool validRequest(enum commands c) {
     case CMD_PAM_CHAUTHTOK:
     case CMD_PAM_FAIL_DELAY:
         return config->auth;
-    case CMD_SETUID:
-        return false;
     case CMD_FORK:
-    case CMD_EXIT:
     case CMD_DAEMON:
         return config->pfork;
     }
@@ -713,9 +843,7 @@ static void control_loop(void) {
         c = (enum commands)msg_getInt(msg);
 
         if (!validRequest(c)) {
-            msg_initResponce(msg, -EPERM);
-            syslog(LOG_NOTICE, "Unknown request '%c'", (char)c);
-            msg_sendmsg(msg, privmand_fd, "control_loop(validRequest)");
+            sendEPERM(msg, "Unknown request");
             continue;
         }
         /* The handler function for a given request does the responce. */
@@ -731,10 +859,11 @@ static void control_loop(void) {
     }
     msg_delete(msg);
 
-    if (readlen < 0)
+    if (readlen < 0 && errno != EBADF)
         boom("recvmsg");
 }
 
+#ifdef HAVE_LIBPAM
 #define PAM_SIMPLE_HANDLER(name, method)                                \
 static void name(message_t *msg) {                                      \
     pamSimpleFunc(msg, method);                                         \
@@ -747,15 +876,18 @@ PAM_SIMPLE_HANDLER(pamSetcred,          pam_setcred)
 PAM_SIMPLE_HANDLER(pamOpenSession,      pam_open_session)
 PAM_SIMPLE_HANDLER(pamCloseSession,     pam_close_session)
 PAM_SIMPLE_HANDLER(pamChauthtok,        pam_chauthtok)
-#ifdef HAVE_PAM_FAIL_DELAY
+#ifdef LO_HAVE_PAM_FAIL_DELAY
 PAM_SIMPLE_HANDLER(pamFailDelay,     (int(*)(pam_handle_t*,int))pam_fail_delay)
 #endif
+#endif /* HAVE_LIBPAM */
 
 void privman_serv_init(void)
 {
     function_map[ CMD_OPEN              ] = openFile;
+    function_map[ CMD_UNLINK            ] = unlinkFile;
     function_map[ CMD_BIND              ] = bindPort;
 
+#ifdef HAVE_LIBPAM
     function_map[ CMD_PAM_START         ] = pamStart;
     function_map[ CMD_PAM_GET_ITEM      ] = pamGetItem;
     function_map[ CMD_PAM_SET_ITEM      ] = pamSetItem;
@@ -763,20 +895,24 @@ void privman_serv_init(void)
     function_map[ CMD_PAM_ACCT_MGMT     ] = pamAcctMgmt;
     function_map[ CMD_PAM_END           ] = pamEnd;
     function_map[ CMD_PAM_SETCRED       ] = pamSetcred;
-    function_map[ CMD_PAM_SETCRED       ] = pamSetcred;
     function_map[ CMD_PAM_OPEN_SESSION  ] = pamOpenSession;
     function_map[ CMD_PAM_CLOSE_SESSION ] = pamCloseSession;
     function_map[ CMD_PAM_CHAUTHTOK     ] = pamChauthtok;
-#ifdef HAVE_PAM_FAIL_DELAY
+#ifdef LO_HAVE_PAM_FAIL_DELAY
     function_map[ CMD_PAM_FAIL_DELAY    ] = pamFailDelay;
 #endif
+#endif /* HAVE_LIBPAM */
 
     function_map[ CMD_FORK              ] = forkProcess;
     function_map[ CMD_EXIT              ] = exitServer;
     function_map[ CMD_DAEMON            ] = daemonProcess;
+    function_map[ CMD_WAIT4             ] = privWait4;
 
     function_map[ CMD_RERUN_AS          ] = rerunAsProcess;
-    function_map[ CMD_EXEC_AS           ] = execAsProcess;
+    function_map[ CMD_RESPAWN_AS        ] = respawnAsProcess;
+
+    function_map[ CMD_CUSTOM_INFO       ] = customInfo;
+    function_map[ CMD_CUSTOM_CAP        ] = customCap;
 
 
     /* Syslog init. */
@@ -788,28 +924,23 @@ void privman_serv_init(void)
 
     /* Two ways we get here
      * 1) CMD_EXIT
-     * 2) CMD_RERUN_AS
+     * 2) CMD_RERUN_AS/RESPAWN_AS
      *
      * RERUN_AS requires magic.  This is a second child of the
-     * privmand server, so we want to call back out of priv_init()
-     */
-    /* If we are not the root parent, exit here so that only
-     * one privmand process returns back to priv_init() in
-     * priv_client.cc to wait on the main child.
-     *
-     * This is either CMD_FORK or CMD_DAEMON.  Either
+     * privmand server, so we want to drop back out to priv_init()
      */
     if (child_pid != 0) { 
-        /* CMD_EXIT */
-        if (p_wait_on_child == false)
-            _exit(0);
-        else {
+        if (p_wait_on_child) {
+            /* Other side closed pipe */
             int status;
             waitpid(child_pid, &status, 0);
             if (WIFEXITED(status))
                 _exit(WEXITSTATUS(status));
-            else
+            else {
                 _exit(EXIT_FAILURE);
+            }
+         } else {
+            _exit(0);
         }
     }
 }
