@@ -5,7 +5,7 @@
  * types.h: Implimentation header.  Contains certain C++ types used
  * by the configuration parser and associated code.
  *
- * $Id: types.h,v 1.14 2002/08/23 19:02:59 dougk Exp $
+ * $Id: types.h,v 1.15 2002/10/27 23:27:48 dougk Exp $
  */
 #ifndef MY_TYPES_H
 #define MY_TYPES_H 1
@@ -23,8 +23,8 @@ struct config_t {
     path_list   open_ro;        /* read-only files              */
     path_list   open_rw;        /* read-write files             */
     path_list   open_ao;        /* append-only files            */
+    path_list   unlink;         /* Files which can be unlinked  */
 
-    path_list   exec_list;      /* run as programs.             */
     user_list   user;           /* run as users.                */
 
     std::string unpriv_user;    /* Unpriviledged user           */
