@@ -7,7 +7,7 @@
  * out.  Its is far more complex than it needs to be (fork?  Please.) in
  * order to test additional Privman functionality.
  *
- * $Id: mycat.c,v 1.5 2002/09/16 20:13:17 dougk Exp $
+ * $Id: mycat.c,v 1.6 2002/10/27 23:27:48 dougk Exp $
  */
 
 #include "../config.h"
@@ -23,9 +23,15 @@ int main(int argc, char *argv[])
 {
     int fd, i, n;
     char buf[4096];
+    char *cwd;
+
+    cwd = getcwd(0,0); /* priv_init does a chdir("/") */
 
     priv_init("mycat");
 
+    if (chdir(cwd) < 0)
+        perror("main(chdir)");
+
     for (i=1; i < argc; ++i) {
         pid_t   child;
 
