@@ -7,7 +7,7 @@
  * out.  Its is far more complex than it needs to be (fork?  Please.) in
  * order to test additional Privman functionality.
  *
- * $Id: mycat2.c,v 1.3 2002/09/16 20:13:17 dougk Exp $
+ * $Id: mycat2.c,v 1.4 2002/10/30 17:30:35 dougk Exp $
  */
 
 #include "../config.h"
@@ -23,10 +23,16 @@ int main(int argc, char *argv[])
 {
     int i, n;
     char buf[4096];
+    char *cwd;
+
+    cwd = getcwd(0,0); /* priv_init does a chdir("/") */
 
     /* Reuse the one config file */
     priv_init("mycat");
 
+    if (chdir(cwd) < 0)
+        perror("main(chdir)");
+
     for (i=1; i < argc; ++i) {
         pid_t   child;
         FILE   *f;
