@@ -6,7 +6,7 @@
  * This program executes "id" as user "joe".  Its simply a small test
  * program for priv_execve.
  *
- * $Id: myexec.c,v 1.4 2002/09/16 20:13:17 dougk Exp $
+ * $Id: myexec.c,v 1.5 2002/10/30 20:49:39 dougk Exp $
  */
 
 #include "../config.h"
@@ -23,9 +23,10 @@ extern char **environ;
 
 int main(void)
 {
+    char *argv[2] = { "id", 0 };
     priv_init("myexec");
 
-    priv_execve(ID_PATH, (char *[]){"id"}, environ, "joe", 0);
+    priv_execve(ID_PATH, argv, environ, "joe", 0);
 
     perror("priv_execve");
 
