@@ -8,7 +8,7 @@
  * program should help demonstrate the use of priv_rerunas(), a less
  * than-obvious Privman method.
  *
- * $Id: myrerun.c,v 1.4 2002/09/16 20:13:17 dougk Exp $
+ * $Id: myrerun.c,v 1.6 2002/10/30 17:34:06 dougk Exp $
  */
 
 #include "../config.h"
@@ -23,15 +23,15 @@
 /* Needs to be global so that rerun_fn and main can both see it. */
 int state = 0;
 
-void rerun_fn(const char * args)
+void rerun_fn(char * const args[])
 {
     /* This function will run before we drop out of priv_init the second
      * through fifth times.  Futz with global state so we can recognize
      * the situation.  args comes from the priv_rerunas call, and is a 
-     * null-terminated string.  Don't try and pass pointers via args,
-     * it won't work.
+     * null-terminated array of strings.  Don't try and pass pointers via
+     * args, it won't work.
      */
-    state = atoi(args);
+    state = atoi(args[0]);
 }
 
 int main(void) {
@@ -42,16 +42,19 @@ int main(void) {
     /* Nothing up my sleeve... */
     printf("state = %d, uid = %d/%d\n", state, getuid(), geteuid());
     if (state < 5) {
-        char arg[20];
+        char **arg;
+        arg = (char**)malloc(sizeof(char *) * 2);
+        arg[0] = malloc(5);
+        arg[1] = 0;
         /* Create a string to pass the state to the next iteration. */
-        snprintf(arg, sizeof(arg)-1, "%d", state + 1);
+        snprintf(arg[0], 4, "%d", state + 1);
 
         /* setuid to user, don't chroot, call rerun_fn(arg) before dropping
          * out of priv_init again.
          */
-        priv_rerunas(rerun_fn, arg, users[state], 0);
-        /* If we get here, something failed.  Probally an unknown user. */
-        exit(EXIT_FAILURE);
+        if ( priv_rerunas(rerun_fn, arg, users[state], NULL, 0) < 0)
+            fprintf(stderr,"priv_rerunas failed.\n");
+        _exit(0);
     }
 
     exit(EXIT_SUCCESS);
